@@ -23,28 +23,35 @@
 ** Author: Peter Eriksson <pen@signum.se>
 */
 
+#include "system.h"
+
 #include <stdio.h>
+#include <stdlib.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <string.h>
 #include <gdbm.h>
+#include <unistd.h>
 #include <sys/stat.h>
 #include "yp.h"
 
 
+#ifndef _PATH_YP
 #define _PATH_YP   "/var/yp"
+#endif
 char *path_ypdb = _PATH_YP;
 
 int debug_flag = 1;
 
 
-static void Gperror(char *str)
+static void Gperror(const char *str)
 {
-    fprintf(stderr, "%s: GDBM Error Code #%d\n", gdbm_errno);
+    fprintf(stderr, "%s: GDBM Error Code #%d\n", str, gdbm_errno);
 }
 
 
-void *ypproc_null_2_svc(void *dummy, struct svc_req *rqstp)
+void *ypproc_null_2_svc(void *dummy,
+			struct svc_req *rqstp)
 {
     static int foo;
 
@@ -58,13 +65,19 @@ void *ypproc_null_2_svc(void *dummy, struct svc_req *rqstp)
 /*
 ** Return 1 if the name is a valid domain name served by us, else 0.
 */
-static int is_valid_domain(char *name)
+static int is_valid_domain(const char *domain)
 {
     char buf[1025];
     struct stat sbuf;
     
+
+    if (strcmp(domain, "binding") == 0 ||
+	strcmp(domain, "..") == 0 ||
+	strcmp(domain, ".") == 0 ||
+	strchr(domain, '/'))
+	return 0;
     
-    sprintf(buf, "%s/%s", path_ypdb, name);
+    sprintf(buf, "%s/%s", path_ypdb, domain);
     if (stat(buf, &sbuf) < 0 || !S_ISDIR(sbuf.st_mode))
 	return 0;
 
@@ -72,7 +85,8 @@ static int is_valid_domain(char *name)
 }
 
 
-bool_t *ypproc_domain_2_svc(domainname *name, struct svc_req *rqstp)
+bool_t *ypproc_domain_2_svc(domainname *name,
+			    struct svc_req *rqstp)
 {
     static bool_t result;
 
@@ -93,7 +107,8 @@ bool_t *ypproc_domain_2_svc(domainname *name, struct svc_req *rqstp)
 }
 
 
-bool_t *ypproc_domain_nonack_2_svc(domainname *name, struct svc_req *rqstp)
+bool_t *ypproc_domain_nonack_2_svc(domainname *name,
+				   struct svc_req *rqstp)
 {
     static bool_t result;
 
@@ -117,12 +132,26 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name, struct svc_req *rqstp)
 }
 
 
-static GDBM_FILE open_database(char *domain, char *map)
+/*
+** Open a GDBM database
+*/
+static GDBM_FILE open_database(const char *domain,
+			       const char *map)
 {
     GDBM_FILE dbp;
     char buf[1025];
 
-    
+
+    /* Better be safe than sorry... */
+    if (strcmp(domain, "binding") == 0 ||
+	strcmp(domain, "..") == 0 ||
+	strcmp(domain, ".") == 0 ||
+	strchr(domain, '/'))
+	return 0;
+
+    if (map[0] == '.' || strchr(map, '/'))
+	return 0;
+
     sprintf(buf, "%s/%s/%s", path_ypdb, domain, map);
     
     dbp = gdbm_open(buf, 0, O_RDONLY, 0, NULL);
@@ -134,10 +163,80 @@ static GDBM_FILE open_database(char *domain, char *map)
 }
 
 
+/*
+** Get a record from a GDBM database.
+*/
+int read_database(GDBM_FILE dbp,
+		  const datum *ikey,
+		  datum *okey,
+		  datum *dval,
+		  int all_flag)
+{
+    int first_flag = 0;
+    datum nkey, ckey;
+    
+    
+    if (ikey == NULL || ikey->dptr == NULL)
+    {
+	ckey = gdbm_firstkey(dbp);
+	first_flag = 1;
+    }
+    else
+	ckey = *ikey;
+    
+    if (ckey.dptr == NULL)
+	return YP_NOKEY;
+
+    while (1)
+    {
+	*dval = gdbm_fetch(dbp, ckey);
+	if (dval->dptr == NULL)
+	{
+	    /* Free key, unless it comes from the caller! */
+	    if (ikey == NULL || ckey.dptr != ikey->dptr)
+		free(ckey.dptr);
+
+	    if (ikey && ikey->dptr != NULL)
+		return YP_NOKEY;
+	    else
+		if (first_flag)
+		    return YP_BADDB;
+		else
+		    return YP_FALSE;
+	}
+	
+	if (all_flag || strncmp(dval->dptr, "YP_", 3) != 0)
+	{
+	    if (okey)
+		*okey = ckey;
+	    else if (ikey == NULL || ikey->dptr != ckey.dptr)
+		free(ckey.dptr);
+	    
+	    return YP_TRUE;
+	}
+
+	/* Free old value */
+	free(dval->dptr);
+	
+	nkey = gdbm_nextkey(dbp, ckey);
+
+	/* Free old key, unless it comes from the caller! */
+	if (ikey == NULL || ckey.dptr != ikey->dptr)
+	    free(ckey.dptr);
+
+	if (ckey.dptr == NULL)
+	    return YP_NOMORE;
+
+	ckey = nkey;
+    }
+}
+
+
 /*
 ** Get the DateTimeModified value for a certain map database
 */
-static unsigned long get_dtm(char *domain, char *map)
+static unsigned long get_dtm(const char *domain,
+			     const char *map)
 {
     struct stat sbuf;
     char buf[1025];
@@ -150,7 +249,11 @@ static unsigned long get_dtm(char *domain, char *map)
 }
 
 
-ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
+/*
+** YP function "MATCH" implementation
+*/
+ypresp_val *ypproc_match_2_svc(ypreq_key *key,
+			       struct svc_req *rqstp)
 {
     static ypresp_val result;
 
@@ -162,7 +265,7 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	fprintf(stderr, "\t\tmapname = \"%s\"\n",
 		key->map);
 	fprintf(stderr, "\t\tkeydat = \"%.*s\"\n",
-		key->key.keydat_len,
+		(int) key->key.keydat_len,
 		key->key.keydat_val);
     }
 
@@ -189,19 +292,14 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	    qdat.dsize = key->key.keydat_len;
 	    qdat.dptr = key->key.keydat_val;
 
-	    rdat = gdbm_fetch(dbp, qdat);
-	    if (rdat.dptr == NULL)
+	    result.stat = read_database(dbp, &qdat, NULL, &rdat, 1);
+
+	    if (result.stat == YP_TRUE)
 	    {
-		if (debug_flag)
-		    Gperror("gdbm_fetch()");
-		result.stat = YP_NOKEY;
-	    }
-	    else
-	    {
-		result.stat = YP_TRUE;
 		result.val.valdat_len = rdat.dsize;
 		result.val.valdat_val = rdat.dptr;
 	    }
+
 	    gdbm_close(dbp);
 	}
     }
@@ -210,7 +308,8 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
     {
 	if (result.stat == YP_TRUE)
 	    fprintf(stderr, "\t-> Value = \"%.*s\"\n",
-		    result.val.valdat_len, result.val.valdat_val);
+		    (int) result.val.valdat_len,
+		    result.val.valdat_val);
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
     }
@@ -218,17 +317,21 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
     return &result;
 }
 
-ypresp_key_val *ypproc_first_2_svc(ypreq_key *key, struct svc_req *rqstp)
+
+
+ypresp_key_val *ypproc_first_2_svc(ypreq_key *key,
+				   struct svc_req *rqstp)
 {
     static ypresp_key_val result;
 
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_match():\n");
+	fprintf(stderr, "ypproc_first():\n");
 	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
 	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
-	fprintf(stderr, "\tkeydat = \"%.*s\"\n", key->key.keydat_len,
+	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
+		(int) key->key.keydat_len,
 		key->key.keydat_val);
     }
 
@@ -259,21 +362,17 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	    result.stat = YP_NOMAP;
 	else
 	{
-	    dkey = gdbm_firstkey(dbp);
-	    if (dkey.dptr == NULL)
-		result.stat = YP_NOKEY;
-	    else
+	    result.stat = read_database(dbp, NULL, &dkey, &dval, 0);
+
+	    if (result.stat == YP_TRUE)
 	    {
-		dval = gdbm_fetch(dbp, dkey);
-		
-		result.stat = YP_TRUE;
-		
 		result.key.keydat_len = dkey.dsize;
 		result.key.keydat_val = dkey.dptr;
 		
 		result.val.valdat_len = dval.dsize;
 		result.val.valdat_val = dval.dptr;
 	    }
+	    
 	    gdbm_close(dbp);
 	}
     }
@@ -281,9 +380,11 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_key *key, struct svc_req *rqstp)
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
-	    fprintf(stderr, "\t-> Key = \"%s\", Value = \"%.*s\"\n",
-		    result.key.keydat_len, result.key.keydat_val,
-		    result.val.valdat_len, result.val.valdat_val);
+	    fprintf(stderr, "\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
+		    (int) result.key.keydat_len,
+		    result.key.keydat_val,
+		    (int) result.val.valdat_len,
+		    result.val.valdat_val);
 		    
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
@@ -293,17 +394,19 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_key *key, struct svc_req *rqstp)
 }
 
 
-ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
+ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
+				  struct svc_req *rqstp)
 {
     static ypresp_key_val result;
 
-
+    
     if (debug_flag)
     {
 	fprintf(stderr, "ypproc_match():\n");
 	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
 	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
-	fprintf(stderr, "\tkeydat = \"%.*s\"\n", key->key.keydat_len,
+	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
+		(int) key->key.keydat_len,
 		key->key.keydat_val);
     }
 
@@ -327,7 +430,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	result.stat = YP_NODOM;
     else
     {
-	datum dkey, dval;
+	datum dkey, dval, okey;
 
 
 	GDBM_FILE dbp = open_database(key->domain, key->map);
@@ -338,36 +441,16 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	    dkey.dsize = key->key.keydat_len;
 	    dkey.dptr  = key->key.keydat_val;
 	    
-	    dkey = gdbm_nextkey(dbp, dkey);
-	    if (dkey.dptr == NULL)
+	    result.stat = read_database(dbp, &dkey, &okey, &dval, 0);
+
+	    if (result.stat == YP_TRUE)
 	    {
-		if (debug_flag)
-		    Gperror("gdbm_nextkey()");
+		result.key.keydat_len = okey.dsize;
+		result.key.keydat_val = okey.dptr;
 		
-		result.stat = YP_NOMORE;
+		result.val.valdat_len = dval.dsize;
+		result.val.valdat_val = dval.dptr;
 	    }
-	    else
-	    {
-		dval = gdbm_fetch(dbp, dkey);
-		if (dval.dptr == NULL)
-		{
-		    if (debug_flag)
-			Gperror("gdbm_fetch()");
-		    
-		    result.stat = YP_BADDB;
-		}
-		else
-		{
-		    result.stat = YP_TRUE;
-		    
-		    result.key.keydat_len = dkey.dsize;
-		    result.key.keydat_val = dkey.dptr;
-		    
-		    result.val.valdat_len = dval.dsize;
-		    result.val.valdat_val = dval.dptr;
-		}
-	    }
-	    
 	    gdbm_close(dbp);
 	}
     }
@@ -375,9 +458,11 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
-	    fprintf(stderr, "\t-> Key = \"%s\", Value = \"%.*s\"\n",
-		    result.key.keydat_len, result.key.keydat_val,
-		    result.val.valdat_len, result.val.valdat_val);
+	    fprintf(stderr, "\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
+		    (int) result.key.keydat_len,
+		    result.key.keydat_val,
+		    (int) result.val.valdat_len,
+		    result.val.valdat_val);
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
     }
@@ -385,7 +470,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
     return &result;
 }
 
-static void print_ypmap_parms(struct ypmap_parms *pp)
+static void print_ypmap_parms(const struct ypmap_parms *pp)
 {
     fprintf(stderr, "\t\tdomain   = \"%s\"\n", pp->domain);
     fprintf(stderr, "\t\tmap      = \"%s\"\n", pp->map);
@@ -394,7 +479,11 @@ static void print_ypmap_parms(struct ypmap_parms *pp)
 }
 
 
-ypresp_xfr *ypproc_xfr_2_svc(ypreq_xfr *xfr, struct svc_req *rqstp)
+/*
+** XXXX Implement Me!
+*/
+ypresp_xfr *ypproc_xfr_2_svc(ypreq_xfr *xfr,
+			     struct svc_req *rqstp)
 {
     static ypresp_xfr result;
 
@@ -415,7 +504,8 @@ ypresp_xfr *ypproc_xfr_2_svc(ypreq_xfr *xfr, struct svc_req *rqstp)
 }
 
 
-void *ypproc_clear_2_svc(void *dummy, struct svc_req *rqstp)
+void *ypproc_clear_2_svc(void *dummy,
+			 struct svc_req *rqstp)
 {
     static int foo;
 
@@ -426,32 +516,118 @@ void *ypproc_clear_2_svc(void *dummy, struct svc_req *rqstp)
 }
 
 
+static int ypall_close(void *data)
+{
+    if (debug_flag)
+	if (data == NULL)
+	    fprintf(stderr, "ypall_close() called with NULL pointer.\n");
+	else
+	    fprintf(stderr, "ypall_close() called.\n");
+	    
+    gdbm_close(*(GDBM_FILE *) data);
+    return 0;
+}
 
-ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
+
+static int ypall_encode(ypresp_key_val *val,
+			void *data)
+{
+    datum dkey, dval, okey;
+    int status;
+
+    if (debug_flag)
+	fprintf(stderr, "ypall_encode() called with key \"%.*s\".\n",
+		(int) val->key.keydat_len,
+		val->key.keydat_val);
+    
+    dkey.dptr = val->key.keydat_val;
+    dkey.dsize = val->key.keydat_len;
+    
+    status = read_database(*(GDBM_FILE *) data, &dkey, &okey, &dval, 0);
+
+    if (status == YP_TRUE)
+    {
+	val->key.keydat_val = okey.dptr;
+	val->key.keydat_len = okey.dsize;
+
+	val->val.valdat_val = dval.dptr;
+	val->val.valdat_len = dval.dsize;
+    }
+    
+    return status;
+}
+
+
+ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
+			     struct svc_req *rqstp)
 {
     static ypresp_all result;
-    
+    extern xdr_ypall_cb_t xdr_ypall_cb;
+
     
     if (debug_flag)
     {
 	fprintf(stderr, "ypproc_all_2_svc():\n");
 	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
 	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
-	fprintf(stderr, "\t-> Request denied (not yet implemented).\n");
     }
+
+    xdr_ypall_cb.u.encode = NULL;
+    xdr_ypall_cb.u.close  = NULL;
+    xdr_ypall_cb.data = NULL;
     
-    result.more = 0;
-    result.ypresp_all_u.val.stat = YP_YPERR;
-    result.ypresp_all_u.val.key.keydat_len = 0;
-    result.ypresp_all_u.val.key.keydat_val = NULL;
-    result.ypresp_all_u.val.val.valdat_len = 0;
-    result.ypresp_all_u.val.val.valdat_val = NULL;
+    result.more = TRUE;
+    
+    if (nokey->map[0] == '\0' || nokey->domain[0] == '\0')
+	result.ypresp_all_u.val.stat = YP_BADARGS;
+    else if (!is_valid_domain(nokey->domain))
+	result.ypresp_all_u.val.stat = YP_NODOM;
+    else
+    {
+	datum dkey, dval;
+	
+	GDBM_FILE dbp = open_database(nokey->domain, nokey->map);
+	if (dbp == NULL)
+	    result.ypresp_all_u.val.stat = YP_NOMAP;
+	else
+	{
+	    result.ypresp_all_u.val.stat = read_database(dbp,
+							 NULL,
+							 &dkey,
+							 &dval,
+							 0);
+
+	    if (result.ypresp_all_u.val.stat == YP_TRUE)
+	    {
+		result.ypresp_all_u.val.key.keydat_len = dkey.dsize;
+		result.ypresp_all_u.val.key.keydat_val = dkey.dptr;
+		
+		result.ypresp_all_u.val.val.valdat_len = dval.dsize;
+		result.ypresp_all_u.val.val.valdat_val = dval.dptr;
+
+		xdr_ypall_cb.u.encode = ypall_encode;
+		xdr_ypall_cb.u.close  = ypall_close;
+		xdr_ypall_cb.data = (void *) &dbp;
+
+		if (debug_flag)
+		    fprintf(stderr, "Returning because YP_TRUE\n");
+		
+		return &result;
+	    }
+
+	    if (debug_flag)
+		fprintf(stderr, "Closing ypall GDBM file.\n");
+	    
+	    gdbm_close(dbp);
+	}
+    }
     
     return &result;
 }
 
 
-ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
+ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey,
+				   struct svc_req *rqstp)
 {
     static ypresp_master result;
 
@@ -512,7 +688,8 @@ ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
 }
 
 
-ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
+ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey,
+				 struct svc_req *rqstp)
 {
     static ypresp_order result;
 
@@ -578,7 +755,8 @@ static void free_maplist(ypmaplist *mlp)
     }
 }
 
-static int add_maplist(ypmaplist **mlhp, char *map)
+static int add_maplist(ypmaplist **mlhp,
+		       char *map)
 {
     ypmaplist *mlp;
 
@@ -599,7 +777,9 @@ static int add_maplist(ypmaplist **mlhp, char *map)
     return 0;
 }
 
-ypresp_maplist *ypproc_maplist_2_svc(domainname *name, struct svc_req *rqstp)
+
+ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
+				     struct svc_req *rqstp)
 {
     static ypresp_maplist result;
     
@@ -635,7 +815,7 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name, struct svc_req *rqstp)
 	{
 	    struct dirent *dep;
 	    
-	    while (dep = readdir(dp))
+	    while ((dep = readdir(dp)) != NULL)
 		if (add_maplist(&result.maps, dep->d_name) < 0)
 		{
 		    result.stat = YP_YPERR;
